/***********************************************************************
;; X interface to the Tags facility for Emacs.
;; Copyright (C) 1992  Brad Mears

;; This file is part of taglist

;; Taglist is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.

;; The GNU General Public License can be obtained via anonymous ftp from
;; prep.ai.mit.edu as pub/gnu/COPYING or pub/gnu/COPYING-2.

;; Taglist is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
***********************************************************************/

/***********************************************************************
 *
 * Name - taglist.c
 *
 * Author - Brad Mears
 *
 * Date Created - 1992-04-13
 * 
 * Purpose
 *
 *    This program provides a point-and-click interface to the tags list
 *    in an emacs session.  This won't be much use without taglist.el.
 *
 **********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/Xlib.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>

/*
 * Include the fallback resources used by XtAppInitialize (created with "ad2c")
 */
static String fallbackResources[] = {
#include "taglist.ad.h"
   NULL
};


/* this gets around a problem I saw on some Sun installations */
#ifndef O_RDONLY
#define O_RDONLY        0
#endif


/*
 * Number of configurable command buttons to create.  The last one always
 * terminates the program.
 */
#define NUMBEROFBUTTONS    6


/*
 * List of values recognized by taglist on input.  These may require
 * additional parameters.
 */
#define TLSIG_RESCAN       1
#define TLSIG_TOGGLE       2


/*
 * define a single structure to hold the application-specific resources
 */
typedef struct {
   int           visibleLines;
   String        labels[NUMBEROFBUTTONS];
   String        commands[NUMBEROFBUTTONS];
} AppResources;

AppResources appResources;

/*
 * define the application-specific resources (see XtGetApplicationResources)
 */
static XtResource resources[] = {
   { "visibleLines",  "VisibleLines", XtRInt, sizeof(XtRInt),
     XtOffset(AppResources *, visibleLines), XtRImmediate,  (caddr_t) 10 },

   { "label1", "Label1", XtRString, sizeof(String),
     XtOffset(AppResources *, labels[0]), XtRString, "" },
   { "command1", "Command1", XtRString, sizeof(String),
     XtOffset(AppResources *, commands[0]), XtRString, "" },

   { "label2", "Label2", XtRString, sizeof(String),
     XtOffset(AppResources *, labels[1]), XtRString, "" },
   { "command2", "Command2", XtRString, sizeof(String),
     XtOffset(AppResources *, commands[1]), XtRString, "" },

   { "label3", "Label3", XtRString, sizeof(String),
     XtOffset(AppResources *, labels[2]), XtRString, "" },
   { "command3", "Command3", XtRString, sizeof(String),
     XtOffset(AppResources *, commands[2]), XtRString, "" },

   { "label4", "Label4", XtRString, sizeof(String),
     XtOffset(AppResources *, labels[3]), XtRString, "" },
   { "command4", "Command4", XtRString, sizeof(String),
     XtOffset(AppResources *, commands[3]), XtRString, "" },

   { "label5", "Label5", XtRString, sizeof(String),
     XtOffset(AppResources *, labels[4]), XtRString, "" },
   { "command5", "Command5", XtRString, sizeof(String),
     XtOffset(AppResources *, commands[4]), XtRString, "" },

   { "label6", "Label6", XtRString, sizeof(String),
     XtOffset(AppResources *, labels[5]), XtRString, "" },
   { "command6", "Command6", XtRString, sizeof(String),
     XtOffset(AppResources *, commands[5]), XtRString, "" },

};

/*
 * Okay, I admit it - I use global variables sometimes.  So shoot me.
 */
static Widget toplevel;                /* upper-most widget                */
static Widget box;                     /* overall geometry manager         */
static Widget port;                    /* viewport for scrolled list       */
static Widget list;                    /* the list itself                  */
static Widget cmdbox;                  /* geometry manager for cmd buttons */
static Widget cmds[NUMBEROFBUTTONS];   /* configurable command buttons     */

static char **taglist;                    /* pointer to list of tags       */
static char  *errlist[] = { NULL, NULL }; /* a taglist to use upon error   */

/***************************************************************************
 ***************************************************************************
                         Utility Routines                
 ***************************************************************************
 **************************************************************************/

/*******************************************************************
 * Calculate the character width and font for a particular Widget.
 *******************************************************************/
static void cal_char_wh( w, char_w, char_h )
Widget   w;
int     *char_w;
int     *char_h;
{
   int         direction, as, ds;
   XCharStruct c_s;
   XFontStruct *fs;

   /*
    * Find the maximum ascent and descent.
    */
   XtVaGetValues(w, XtNfont, &fs, NULL);
   XTextExtents(fs, "Xq", 2, &direction, &as, &ds, &c_s);
   *char_h = (int) as + ds + 2;

   /*
    * Find the maximum width
    */
   XTextExtents(fs, "X", 1, &direction, &as, &ds, &c_s);
   *char_w = (int) c_s.width;
}



/****************************************************************************
 * Given a viewport widget that contains a list widget, adjust the size
 * of the viewport so that the scrollbars don't overwrite the list.
 * Returns non-zero if an error occurred.
 ****************************************************************************/
static void tailor_scrolled_list(port, list)
Widget port;
Widget list;
{
   Widget    vertical;                   /* Widget ID of vert.  scrollbar */
   Dimension thickness;                  /* scrollbar thickness           */
   int       portHeight;                 /* height of viewport widget     */
   int       portWidth;                  /* width  of viewport widget     */
   int       longest;                    /* longest entry in the list     */
   int       char_w;                     /* max width of any character    */
   int       char_h;                     /* max height of any character   */
   int       lines;                      /* # of lines to display         */
   int       count;                      /* # of entries in current list  */

   /*
    * query the list widget for its current info
    */
   cal_char_wh(list, &char_w, &char_h);         /* find font dimensions */
   XtVaGetValues(list, XtNnumberStrings, &count,
                       XtNlongest,       &longest, 
                       NULL );

   /*
    * Compute the desired height of the viewport
    */
   if (count < appResources.visibleLines) {
      portHeight = char_h * count + 4;
   } else {
      portHeight = char_h * appResources.visibleLines + 4;
   }

   /*
    * Compute the desired width of the viewport
    */      
   vertical = XtNameToWidget(port,"vertical");
   if ( vertical != NULL ) {
      XtVaGetValues( vertical, XtNwidth, &thickness, NULL );
      portWidth = longest + char_w + thickness;
   } else {
      portWidth = longest + char_w;
   }

   /*
    * Set the height & width of the viewport
    */
   XtVaSetValues(port, XtNheight, portHeight,
                       XtNwidth,  portWidth, 
                       NULL );
}


/***************************************************************************
 ***************************************************************************
                          Callbacks                
 ***************************************************************************
 **************************************************************************/

/******************************************************************
 * Callback for the configurable command buttons.  The last button
 * is ALWAYS exit.  The button number is passed as client_data.
 ******************************************************************/
/*ARGSUSED*/
static void cmd_cb( w, client_data, call_data )
Widget  w;
caddr_t client_data; 
caddr_t call_data;
{
   int buttonNo = (int) client_data;

   if ( buttonNo == (NUMBEROFBUTTONS-1) )
      exit(0);

   fputs(appResources.commands[buttonNo],stdout);
   fflush(stdout);
}


/**************************************************************************
 * Callback for the list widget.  Insert the tag name into a (find-tag)
 * command and write it out.  The selected item in the list is passed
 * as call_data.
 **************************************************************************/
/*ARGSUSED*/
static void list_cb( w, client_data, call_data )
Widget  w;
caddr_t client_data;
caddr_t call_data;
{
   XawListReturnStruct *returnStruct = (XawListReturnStruct *) call_data;

   fprintf(stdout,"(find-tag \"%s\")",returnStruct->string);
   fflush(stdout);
}



/***************************************************************************
 ***************************************************************************
                         Everything else                
 ***************************************************************************
 **************************************************************************/

/******************************************************************
 * Create a simple box to hold everything
 ******************************************************************/
static void create_box( parent )
Widget parent;
{
   static Arg args[] = {
      { XtNorientation, XtorientHorizontal },
   };

   box = XtCreateManagedWidget(
              "box",                /* arbitrary widget name */
              boxWidgetClass,       /* widget class */
              parent,               /* parent widget */
              args,                 /* argument list */
              XtNumber(args)        /* arg list size */
              );
}


/******************************************************************
 * Create a simple box to hold the command buttons
 ******************************************************************/
static void create_buttons( parent )
Widget parent;
{
   int  i;                                      /* loop counter            */
   Dimension  thisWidth;                        /* width of current button */
   Dimension  maxWidth = 0;                     /* maximum button width    */
   static Arg args[] = {                        /* args for box widget     */
      { XtNorientation, XtorientVertical },
      { XtNborderWidth, (XtArgVal) 0 },
   };

   /*
    * Create a box widget to hold the commands
    */
   cmdbox = XtCreateManagedWidget(
              "cmdbox",             /* widget name */
              boxWidgetClass,       /* widget class */
              parent,               /* parent widget */
              args,                 /* argument list */
              XtNumber(args)        /* arg list size */
              );

   /*
    * Create each of the buttons
    */
   maxWidth = 0;
   for (i=0; i<NUMBEROFBUTTONS; i++) {
      cmds[i] = XtCreateManagedWidget(
              appResources.labels[i],  /* widget name */
              commandWidgetClass,      /* widget class */
              cmdbox,                  /* parent widget */
              NULL,                    /* argument list */
              0                        /* arg list size */
              );

      XtVaGetValues( cmds[i], XtNwidth, &thisWidth, NULL );
      if (maxWidth < thisWidth)
	 maxWidth = thisWidth;

      XtAddCallback( cmds[i], XtNcallback, cmd_cb, i );
   }

   /*
    * Set all the buttons to the same width
    */
   for (i=0; i<NUMBEROFBUTTONS; i++) {
      XtVaSetValues( cmds[i], XtNwidth, maxWidth, NULL );
   }
}


/****************************************************************************
 * Create a list widget to hold the list of tags.  Put the list in a
 * viewport.
 ****************************************************************************/
static void create_list( parent, taglist )
Widget parent;
char **taglist;
{
   Arg   args[10];                        /* Xt argument list                */
   int   n;                               /* number of entries in Args       */
   static XtCallbackRec callbacks[] = {   /* callback list for list widget   */
      { list_cb, (caddr_t) NULL },
      { (XtCallbackProc) NULL, (caddr_t) NULL }
   };


   /*
    * Create the viewport widget 
    */
   n = 0;
   XtSetArg(args[n], XtNborderWidth,  1);                        n++;
   XtSetArg(args[n], XtNallowVert,    True);                     n++;
   XtSetArg(args[n], XtNallowHoriz,   False);                    n++;
   XtSetArg(args[n], XtNforceBars,    True);                     n++;
   port = XtCreateManagedWidget(
				"viewport", 
				viewportWidgetClass, 
				parent,
				args, 
				n
				);
   
   /*
    * Create the list widget
    */
   n = 0;
   XtSetArg( args[n], XtNdefaultColumns, 1 );                         n++;
   XtSetArg( args[n], XtNforceColumns,   TRUE );                      n++;
   XtSetArg( args[n], XtNlist,           taglist);                    n++;
   XtSetArg( args[n], XtNcallback,       callbacks);                  n++;
   list = XtCreateManagedWidget(
              "list",               /* arbitrary widget name */
              listWidgetClass,      /* widget class */
              port,                 /* parent widget */
              args,                 /* argument list */
              n                     /* arg list size */
              );

   /*
    * Determine how high and wide the viewport should be by looking at the 
    * size of the list widget. 
    */ 
   tailor_scrolled_list(port,list);
}


/****************************************************************************
 * Read the contents of filename into memory and construct an array of
 * pointers to each entry.  This replaces all the newlines with nuls.
 ****************************************************************************/
static char **GetFileContents( filename )
char *filename;
{
   struct stat buf;               /* results of stat system call       */
   int   fd;                      /* file descriptor for input file    */
   int   n_read;                  /* bytes transferred during read()   */
   int   i;                       /* loop counter                      */
   int   count;                   /* # of lines in input file          */
   char *p;                       /* a different kind of loop counter  */
   char *contents;                /* pointer to raw data (all of file) */
   char *LastByte;                /* address of end of contents        */
   char **taglist;                /* array of pointers to tag entries  */

   /*
    * Find the total size of the file and allocate enough memory to hold
    * its contents.
    */
   if (stat(filename,&buf)) {
      errlist[0] = "TAGLIST file does not exist";
      return errlist;
   }

   contents = malloc(buf.st_size + 1);
   if (contents == NULL) {
      fputs("malloc failed\n",stderr);
      exit(1);
   }

   /*
    * Read the contents of the file
    */
   fd = open(filename, O_RDONLY, 0);
   if ( fd == -1 ) {
      errlist[0] = "Can't open TAGLIST file";
      return errlist;
   }

   n_read = read( fd, contents, buf.st_size );
   if (n_read == -1) {
      errlist[0] = "Can't read from TAGLIST file";
      return errlist;
   }

   close( fd );

   /*
    * Find the number of individual lines, get rid of any newlines
    */
   count     = 0;
   LastByte  = contents + buf.st_size;
   *LastByte = '\0';
   for (p=contents; (p<LastByte) && (p!=NULL); count++) {
      p = strchr(p,'\n');
      if (p != NULL)
	 *(p++) = '\0';
   }

   /*
    * Allocate a chunk of memory to hold a list of pointers to each entry.
    */
   taglist = (char **) calloc( (sizeof (char *)) ,count+1);
   if (taglist == NULL) {
      fputs("calloc failed",stderr);
      exit(1);
   }

   /*
    * Put the address of each entry in taglist.
    */
   for (i=0,p=contents; i<count; i++) {
      taglist[i] = p;
      p = p + strlen(p) + 1;
   }
   taglist[count] = NULL;    /* NULL terminate it so the List widget doesn't */
			     /* puke */
   return taglist;
}



/****************************************************************************
 * Process inputs from stdin.  This is how emacs communicates with this
 * program.  The list of valid requests is defined by the macros TLSIG_*.
 * Normally a function that handles input events looks at its parameter list
 * for the file id to process.  In this case, I know that I am only 
 * interested in stdin and therefore ignore the fid.
 ****************************************************************************/
void Read_Stdin( client_data, fid, id )
XtPointer client_data;
int *fid;
XtInputId *id;
{
   static char buf[BUFSIZ];             /* buffer for input string          */
   static char dataValue[BUFSIZ];       /* buffer for input arguments       */
   static int isVisible = 1;            /* is taglist displayed?            */
   char **oldlist;                      /* temporary marker for old list    */
   int    fieldCount;                   /* # of fields that were input      */
   int    cmdValue;                     /* command ID to be performed       */

   /*
    * Read input from stdin.  We know it is ready to be read or else this
    * routine wouldn't have been called.  The value in fieldCount is used
    * to validate the format of the input data.  If the value in fieldCount
    * is inappropriate for the desired function, don't do anything.
    */
   fgets(buf,BUFSIZ,stdin);
   fieldCount = sscanf(buf,"%d %s\n",&cmdValue,dataValue);

   /*
    * What am I supposed to do?
    */
   switch (cmdValue) {
   case TLSIG_RESCAN:                         /* rescan a new TAGLIST file */
      if (fieldCount == 2) { 
	 oldlist = taglist;
	 taglist = GetFileContents(dataValue);
         if (oldlist != errlist) {              /* errlist wasn't malloc'ed! */
	    free(*oldlist);
	    free(oldlist);
         }

         XtUnmanageChild(port);
	 XawListChange(list, taglist, 0, 0, True);
         tailor_scrolled_list(port,list);
         XtManageChild(port);
      } else {
         fputs("Invalid command format\n",stdout);
      }
      break;
   case TLSIG_TOGGLE:                             /* toggle the visibility */
      if (fieldCount == 1) {
	 if (isVisible)
	    XtUnmapWidget(toplevel);
	 else
	    XtMapWidget(toplevel);
	 isVisible = !isVisible;
      } else {
         fputs("Invalid command format\n",stdout);
      }
      break;
   default:
      fprintf(stdout,"Unknown command %s\n",buf);
   }
}


/****************************************************************************
 * The main routine for taglist
 ****************************************************************************/
int main( argc, argv )
int argc;
char **argv;
{
   XtAppContext appContext;
   char *filename;

   /*
    * Initialize X
    */
   toplevel = XtAppInitialize(&appContext, "taglist", NULL, 0,
                              &argc,argv,fallbackResources,NULL,0);

   XtGetApplicationResources(toplevel,(XtPointer)&appResources,
                             resources,XtNumber(resources),NULL,0);


   /*
    * Find the file that has the taglist and read it into memory
    */
   if (argc > 1)
      filename = argv[1];
   else
      filename = "TAGLIST";

   taglist = GetFileContents(filename);

   /*
    * Create all the child widgets
    */
   create_box(toplevel);
   create_list(box, taglist);
   create_buttons(box);

   XtAppAddInput(appContext, fileno(stdin), XtInputReadMask, Read_Stdin, NULL);
   XtRealizeWidget(toplevel);
   XtAppMainLoop(appContext);
   return 0;
}
